#!/usr/bin/env python3
import json

from sysrepo_api import sysrepo_api3 as api
import vppapi_if.vppapi_if as vppif
import ucpeif.wireguardvpnif as wireguardif
from netaddr import IPNetwork, IPAddress
from pyroute2 import IPDB
import ipaddress

class SysrepoWireGuardConcentrator(api.SysrepoAPI):
    def __init__(self,
                 app_name="wg-concentrator",
                 logger=None,
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 journal_log=True,
                 namespace="/ucpe:config/wg-concentrators:wg-concentrators",
                 indexes={"wg-concentrator": "name"},
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            logger=logger,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            journal_log=journal_log,
            namespace=namespace,
            indexes=indexes,
            run_as_subscriber=run_as_subscriber,
            priority=priority
        )

        self._wgc_interfaces = []
    
    def get_startup_config(self, full_config=False, full_path=True):
        return self.get_config_datastore("startup", full_config=full_config, full_path=full_path)
    
    def get_running_config(self, full_config=False, full_path=True):
        return self.get_config_datastore("running", full_config=full_config, full_path=full_path)
    
    def get_network_mode(self):
        return self.get_item("/ucpe:config/system:system/network-mode")
    
    def is_vpp_running(self):
        if self.get_network_mode() != 'vpp':
            self.logger.error('uCPE is not in VPP mode')
            return
    
    def get_wireguard_concentrators(self, startup=False):
        config_indexed = {}
        sys_conf = self.get_startup_config(full_path=False) if startup else self.get_running_config(full_path=False)
        wg_concentrators = sys_conf.get('wg-concentrator', [])
        for concentrator_info in wg_concentrators:
            config_indexed[concentrator_info['name']] = concentrator_info
        return config_indexed
    
    def apply_startup_config(self):
        self.is_vpp_running()
        
        self.logger.debug('Apply STARTUP config for Wireguard Concentrator')
        config_indexed = self.get_wireguard_concentrators(startup=True)
        for conc_int_name, config in sorted(config_indexed.items()):
            self.logger.error(f"startup conc_int_name {conc_int_name}")
            concentrator_endpoint = config.get('concentrator-endpoint', '')
            concentrator_port = config.get('concentrator-port', None)
            concentrator_password = config.get('concentrator-password', '')
            concentrator_tunnel_ip = config.get('ipv4-prefix', '')
            is_enabled = config.get('enable', False)
            
            prerequisites = (is_enabled, concentrator_endpoint, concentrator_password, concentrator_port, concentrator_tunnel_ip)
            
            if [x for x in prerequisites if x is None or x == '' or x is False]:
                self.logger.error("ON RETURN")
                continue

            wgc_if = wireguardif.WireGuardVPNInterface(conc_int_name, logger=self.logger, l_port=concentrator_port,
                                                l_endpoint=concentrator_endpoint, enable=is_enabled, is_concentrator=True)
            wgc_if.ip4_addr = ipaddress.IPv4Interface(str(concentrator_tunnel_ip))
            self._wgc_interfaces.append(wgc_if)

        return True
        
    def apply_running_config(self, name, config, change_set):
        self.logger.debug('Apply RUNNING config for Wireguard Concentrator')
        self.logger.error(f"Change_set: {change_set}")
        
        c_enabled = False
        concentrator_endpoint = None
        concentrator_port = None
        concentrator_password = None
        ipv4_prefix = None

        self.logger.error("Config {}".format(config))
        if config:
            c_enabled = config.get('enable', False)
            ipv4_prefix = config.get('ipv4-prefix', None)
            ifname = config.get('name', None)
            concentrator_endpoint = config.get('concentrator-endpoint', None)
            concentrator_password = config.get('concentrator-password', None)
            concentrator_port = config.get('concentrator-port', None)

            self.logger.error(f"PESH1 {ifname} {c_enabled} {ipv4_prefix} {concentrator_endpoint} {concentrator_password} {concentrator_port}")
        
        if change_set:
            self.logger.debug("APPLYING RUNNING FOR WIREGUARD INTERFACE - {}".format(name))
            # self.logger.debug('change_set {} ' .format(change_set))
            ch_set_keys = change_set.keys()
            self.logger.debug('Keys : {}'.format(ch_set_keys))

            wg_if = next((a for a in self._wgc_interfaces if a.name == name), None)
            self.logger.error(f"wg_if {wg_if}")

            if 'name' in ch_set_keys:
                if change_set.get('name', None):
                    if change_set.get('name').get('operation', None) == 'CREATED':
                        if wg_if:
                            self.logger.error("There is WireGuard interface object for {} ?!?!".format(name))
                        else:
                            wg_if = wireguardif.WireGuardVPNInterface(name, logger=self.logger, is_concentrator=True)
                            self._wgc_interfaces.append(wg_if)
                    elif change_set.get('name').get('operation', None) == 'DELETED':
                        if wg_if:
                            self._wgc_interfaces.remove(wg_if)
                            del wg_if
                        return
            
            if 'concentrator-port' in ch_set_keys:
                oper = change_set.get('concentrator-port').get('operation', None )
                local_port = None

                self.logger.error('Local port is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    local_port = change_set.get('concentrator-port').get('new_value', None)

                wg_if.l_port = int(local_port)
                
            if 'concentrator-endpoint' in ch_set_keys:
                oper = change_set.get('concentrator-endpoint').get('operation', None )
                local_endpoint = None

                self.logger.error('Local endpoint is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    local_endpoint = change_set.get('concentrator-endpoint').get('new_value', None)

                l_local, l_localv6 = self._get_local_wireguard_endpoint_ip(local_endpoint)
                
                if l_local:
                    wg_if.l_endpoint = l_local
                    self.logger.error('Local endpoint configured {}' .format(l_local))
                if l_localv6:
                    wg_if.l_endpoint_v6 = l_localv6
                    self.logger.error('Local endpoint configured {}' .format(l_localv6))
                    
            if 'enable' in ch_set_keys:
                oper = change_set.get('enable').get('operation', None)
                c_enable = False

                self.logger.error('State is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    if change_set.get('enable').get('new_value', False) == 'true' or \
                            change_set.get('enable').get('new_value', False) == 'True' or \
                            change_set.get('enable').get('new_value', False) == True:
                        c_enable = True
                        # self.__add_del_ip_pass_tap_intf_if_wg_enabled(ifname, 1)
                    else:
                        # self.__add_del_ip_pass_tap_intf_if_wg_enabled(ifname, 0)
                        pass
                    
                wg_if.enable = c_enable
                
            if 'ipv4-prefix' in ch_set_keys:
                oper = change_set.get('ipv4-prefix').get('operation', None)
                c_ip4_addr = None

                self.logger.error('IPv4 prefix is changed - operation {}' .format(oper))
                if oper == 'MODIFIED' or oper == 'CREATED':
                    c_ip4_addr = change_set.get('ipv4-prefix').get('new_value', None)
                    if c_ip4_addr:
                        c_ip4_addr = ipaddress.IPv4Interface(str(c_ip4_addr))
                    
                wg_if.ip4_addr = c_ip4_addr     
    
    def post_apply_config(self, change_set):
        self.logger.error(f"POST APPLY- change_set {change_set}")
        config_indexed = self.get_wireguard_concentrators(startup=True)
        self.logger.error(f"STARTUP {config_indexed}")

    
    def apply_config(self, change_set=None, apply_startup=False):
        self.logger.debug("Apply Wireguard Concentrator configuration")

        if apply_startup:
            self.apply_startup_config()
        else:
            config_indexed = self.get_wireguard_concentrators(startup=False)
            if not change_set:
                self.logger.error(" There are no changes in change set")
                return config_indexed
            
            concentrator_changes = {}
            for change in change_set:
                if not 'operation' in change:
                    self.logger.error('Malformed change set')
                    continue
                
                name = str(change['xpath'])
                name = name[name.find("wg-concentrator[name='") + len("wg-concentrator[name='"):]
                name = name[:name.find("]") - 1]
                if name not in concentrator_changes:
                    concentrator_changes[name] = {}
                
                concentrator_changes[name][change['xpath'][(change['xpath'].find(']/') + len(']/')):]] = change
            
            for name, xpath_dict in concentrator_changes.items():
                if name in config_indexed:
                        self.apply_running_config(name, config_indexed[name], xpath_dict)
                else:
                    self.apply_running_config(name, None, xpath_dict)

        return True

    # params to set: concentrator-endpoint, concentrator-password, ipv4-prefix, enable, static-range, dynamic-range, domain-name, static-routes;
    def get_wireguard_concentrator_interfaces(self):
        return self.get_running_config(full_path=False)
            
    # params to set: concentrator-endpoint, concentrator-password, ipv4-prefix, enable, static-ip-range, dynamic-ip-range, domain-name, static-routes;
    def set_wireguard_concentrator_param(self, ifname, param, value):
        self.set_item("{}/wg-concentrator[name='{}']/{}".format(self.namespace, ifname, param), value)

    def unset_wireguard_concentrator_param(self, ifname, param):
        self.delete_item("{}/wg-concentrator[name='{}']/{}".format(self.namespace, ifname, param))
        
    def del_wireguard_concentrator(self, name):
        self.delete_item("{}/wg-concentrator[name='{}']".format(self.namespace, name))

    def get_wireguard_concentrator_param(self, name, param, default=None):
        try:
            return self.get_item("{}/wg-concentrator[name='{}']/{}".format(self.namespace, name, param))
        except:
            return default
        
    def get_wireguard_concentrator_params(self, name, default=None):
        try:
            return self.session.get_data("{}/wg-concentrator[name='{}']".format(self.namespace, name))
        except:
            return default
        
    def set_wireguard_concentrator_static_clients(self, ifname, hostname, param, value):
        return self.set_item("{}/wg-concentrator[name='{}']/static-clients[hostname='{}']/{}".format(self.namespace, ifname, hostname, param), value)
    
    def unset_wireguard_concentrator_static_clients(self, ifname, hostname, param):
        return self.delete_item("{}/wg-concentrator[name='{}']/static-clients[hostname='{}']/{}".format(self.namespace, ifname, hostname, param))
    
    def del_wireguard_concentrator_static_clients(self, ifname, hostname):
        return self.delete_item("{}/wg-concentrator[name='{}']/static-clients[hostname='{}']".format(self.namespace, ifname, hostname))
    
    def get_wireguard_concentrator_static_client(self, name, hostname, default=None):
        try:
            return self.get_item("{}/wg-concentrator[name='{}']/static-clients[hostname='{}']".format(self.namespace, name, hostname))
        except:
            return default
    
    def get_wireguard_concentrator_static_clients(self, name, default=None):
        try:
            return self.session.get_data("{}/wg-concentrator[name='{}']/static-clients".format(self.namespace, name))
        except:
            return default
    
    def set_wireguard_concentrator_external_peers(self, ifname, endpoint, param, value):
        return self.set_item("{}/wg-concentrator[name='{}']/external-peers[peer-endpoint='{}']/{}".format(self.namespace, ifname, endpoint, param), value)
    
    def unset_wireguard_concentrator_external_peers(self, ifname, endpoint, param):
        return self.delete_item("{}/wg-concentrator[name='{}']/external-peers[peer-endpoint='{}']/{}".format(self.namespace, ifname, endpoint, param))
    
    def del_wireguard_concentrator_external_peers(self, ifname, endpoint):
        return self.delete_item("{}/wg-concentrator[name='{}']/external-peers[peer-endpoint='{}']".format(self.namespace, ifname, endpoint))
    
    def get_wireguard_concentrator_external_peer(self, name, endpoint, default=None):
        try:
            return self.session.get_data("{}/wg-concentrator[name='{}']/external-peers[peer-endpoint='{}']".format(self.namespace, name, endpoint))
        except:
            return default
        
    def get_wireguard_concentrator_external_peers(self, name, default=None):
        try:
            return self.get_item("{}/wg-concentrator[name='{}']/external-peers".format(self.namespace, name))
        except:
            return default
        
    def _get_local_wireguard_endpoint_ip(self, endpoint):
        n_local = ipaddress.ip_address(endpoint) if isinstance(ipaddress.ip_address(endpoint), ipaddress.IPv4Address) else None
        n_localv6 = ipaddress.ip_address(endpoint) if isinstance(ipaddress.ip_address(endpoint), ipaddress.IPv6Address) else None

        return (n_local, n_localv6)
